'use strict';
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode = require("vscode");
const request = require("request-promise");
const entityDecode = require("decode-html");
const crplData = require("./crpl-data.json");
const helpers_1 = require("./helpers");
const lint_1 = require("./lint");
const documents = new Map();
const diagnostics = vscode.languages.createDiagnosticCollection('crpl');
const crplSelector = { language: 'crpl', scheme: 'file' };
const tokenPattern = /(?:(?:<-|->|-\?|--|@|\$)\w+|:[a-z]\w*)\b|-?\b\d+(?:\.\d*)?\b|(?:<-!|->!|-\?!|--\?)(?=\s|$)|\w+|\S/gi;
// note: this should be the same as language-configuration.json.wordPattern except with the addition of |\S
const symbolPatterns = new Map([
    ['read', /^<-\w+\b$/],
    ['write', /^->\w+\b$/],
    ['exists', /^-\?\w+\b$/],
    ['delete', /^--\w+\b$/],
    ['define', /^\$\w+\b$/],
    ['call', /^@\w+\b$/],
    ['func', /^:\w+\b$/],
    ['refread', /^<-!$/],
    ['refwrite', /^->!$/],
    ['refexists', /^-\?!$/],
    ['refdelete', /^--\?$/],
]);
const prefixPatterns = [
    /^(<-|->|--|-\?|\$)(?=\w)/,
    /^(:|@)(?=\w)/
];
const docPattern = /=====\s*(.*)\s*=====\s*(.*\s*)*?.*Arguments.*\^\s*\|\s*(.*?)\s*\|\s*(.*?)\s*\|(.*?)\|(.*\s*)*?===\s*Description.*\s*((.*\s*)*)/;
class RichDoc {
    constructor(docref) {
        // tokens.flat and tokens.tree should be in warped order, tokens.lines can't be
        this.funcs = new Map();
        this.vars = new Map();
        if (docref.getText) {
            this.doc = docref;
        }
        else {
            this.doc = vscode.workspace.textDocuments.find(doc => doc.uri === docref);
        }
        this.tokens = { flat: [], lines: [], tree: new helpers_1.ParseTree([RichDoc.startToken()]) };
    }
    static startToken() {
        return {
            token: '',
            id: 'start',
            range: {
                start: { line: 0, character: 0 },
                end: { line: 0, character: 0 }
            },
            meta: {}
        };
    }
    tokenize() {
        try {
            let tokens = this.tokens.flat = [];
            let lines = this.tokens.lines = [];
            let tree = this.tokens.tree = new helpers_1.ParseTree([RichDoc.startToken()]);
            this.vars.clear();
            this.funcs.clear();
            const spec = RichDoc.stackSpec;
            const doc = this.doc;
            const ctx = {
                doc: this,
                notLinting: false,
                disabledLints: new Set(),
                tokens, lines, tree,
                branch: tree,
                parametric: true,
                mode: helpers_1.ParseMode.normal,
                line: -1,
                exeStack: [],
                warpStack: []
            };
            const lint = new lint_1.Linter(ctx);
            function addToken(rToken, line) {
                tokens.push(rToken);
                ctx.branch.push(rToken);
                lines[line].push(rToken);
            }
            doc.getText().split('\n').forEach((text, line) => {
                lines.push([]);
                // Here, replace is used as "for each match, do..."
                ctx.line = line;
                text.replace(tokenPattern, (token, offset) => {
                    let range = new vscode.Range(line, offset, line, offset + token.length);
                    if (ctx.mode === helpers_1.ParseMode.normal) {
                        let id;
                        let wiki = true;
                        let error;
                        let rToken = { token, error, range, meta: { wiki } };
                        if (token === '#') {
                            ctx.mode = helpers_1.ParseMode.comment;
                            id = 'comment';
                            wiki = false;
                        }
                        else {
                            if (ctx.parametric) {
                                if (ctx.exeStack.length === 0) {
                                    if (token[0] === '$') {
                                        id = 'define';
                                        rToken.meta.var = token.toLowerCase().slice(1);
                                        rToken.meta.notLinting = ctx.notLinting;
                                        rToken.meta.disabledLints = new Set(ctx.disabledLints);
                                        this.getVar(rToken.meta.var).define.push(rToken);
                                    }
                                    else {
                                        ctx.parametric = false;
                                    }
                                }
                                else if (ctx.exeStack.length === 1 && ctx.exeStack[0].id === 'define') {
                                    wiki = false;
                                    if (token === ':') {
                                        id = ':';
                                    }
                                    else {
                                        error = lint.getError('define', ':');
                                        ctx.parametric = false;
                                    }
                                }
                                else if (ctx.exeStack.length === 1 && ctx.exeStack[0].id === ':') {
                                    wiki = false;
                                    if (RichDoc.valuePattern.test(token)) {
                                        id = 'value';
                                        wiki = false;
                                    }
                                    else {
                                        error = lint.getError('define', 'value');
                                        ctx.parametric = false;
                                    }
                                    if (token === '"') {
                                        ctx.mode = helpers_1.ParseMode.string;
                                    }
                                }
                                else {
                                    ctx.parametric = false;
                                }
                                if (error) {
                                    ctx.exeStack.pop();
                                }
                            }
                            let lowerToken = token.toLowerCase();
                            if (!ctx.parametric) {
                                if (token[0] === '$') {
                                    id = 'define';
                                    // need to include these here since checkVariables takes place later
                                    rToken.meta.notLinting = ctx.notLinting;
                                    rToken.meta.disabledLints = new Set(ctx.disabledLints);
                                    error = lint.getError('definestart');
                                }
                                else if (token === '"') {
                                    ctx.mode = helpers_1.ParseMode.string;
                                    id = token;
                                    wiki = false;
                                }
                                else if (token === '(' || token === ')') {
                                    id = token;
                                    wiki = false;
                                }
                                else if (crplData.words.indexOf(lowerToken) > -1) {
                                    id = lowerToken;
                                }
                                else if (crplData.unitConstants.hasOwnProperty(lowerToken)) {
                                    id = lowerToken;
                                }
                                else if (/^-?\d+(.\d*)?$/.test(token)) {
                                    wiki = false;
                                }
                                else {
                                    symbolPatterns.forEach((re, tokenType) => {
                                        if (re.test(token)) {
                                            id = tokenType;
                                            // need to include these here since checkVariables and checkFunctions take place later
                                            rToken.meta.notLinting = ctx.notLinting;
                                            rToken.meta.disabledLints = new Set(ctx.disabledLints);
                                            if (RichDoc.keyofVarTracker.test(id)) {
                                                let name = lowerToken.slice(2);
                                                rToken.meta.var = name;
                                                this.getVar(name)[tokenType].push(rToken);
                                            }
                                            else if (RichDoc.keyofFuncTracker.test(id)) {
                                                let name = lowerToken.slice(1);
                                                rToken.meta.func = name;
                                                this.getFunc(name)[tokenType].push(rToken);
                                            }
                                        }
                                    });
                                    error = id ? undefined : lint.getError('unknown');
                                    wiki = !!id;
                                }
                            }
                        }
                        rToken.id = id;
                        rToken.meta.wiki = wiki;
                        rToken.error = error;
                        lint.checkUnnest(rToken);
                        lint.checkWarp(rToken);
                        addToken(rToken, line);
                        lint.checkUnwarp(rToken);
                        lint.checkNest(rToken);
                        lint.checkIJK(rToken);
                    }
                    else if (ctx.mode === helpers_1.ParseMode.string) {
                        if (ctx.parametric && ctx.exeStack.length === 0) {
                            // awful hacks in order to preserve a simple RichDoc.stackSpec schema
                            let lastToken = ctx.branch.pop();
                            lastToken.id = '"';
                            ctx.exeStack.push(lastToken);
                            let newBranch = new helpers_1.ParseTree();
                            newBranch.start = lastToken;
                            ctx.branch.push(lastToken, newBranch);
                            ctx.branch = newBranch;
                        }
                        if (token === '"') {
                            ctx.exeStack.pop();
                            ctx.branch = ctx.branch.parent || ctx.branch;
                            addToken({ token, range, id: '"', meta: {} }, line);
                            ctx.mode = helpers_1.ParseMode.normal;
                        }
                        else {
                            addToken({ token, range, id: 'string', meta: {} }, line);
                        }
                    }
                    else if (ctx.mode === helpers_1.ParseMode.comment) {
                        let rToken = { token, range, id: '#' + token, meta: {} };
                        lint.checkDirectives(rToken);
                        addToken(rToken, line);
                    }
                    // this.printTokens()
                    return '';
                });
                if (ctx.mode === helpers_1.ParseMode.comment) {
                    ctx.exeStack.pop();
                    ctx.branch = ctx.branch.parent || ctx.branch;
                    let rToken = {
                        token: '', id: 'endcomment', meta: {},
                        range: new vscode.Range(line, text.length, line, text.length)
                    };
                    lint.checkDirectives(rToken);
                    addToken(rToken, line);
                    ctx.mode = helpers_1.ParseMode.normal;
                }
            });
            // diagnose unresolved blocks
            ctx.exeStack.forEach(rToken => {
                if (rToken.id === 'func') {
                    rToken.error = ((ctx.exeStack.length > 1) || undefined) && lint.getError('enddelim', 'func');
                    return;
                }
                if (rToken.error) {
                    return;
                }
                rToken.parent = rToken.parent;
                let startOffset = spec.unmatched.get(rToken.id);
                let errorToken = rToken.parent[rToken.parent.indexOf(rToken) - startOffset];
                if (errorToken.error) {
                    return;
                }
                errorToken.error = lint.getError('startdelim', rToken.id, errorToken);
            });
            this.checkVariables();
            this.checkFunctions();
            // this.printTokens()
        }
        catch (err) {
            console.error(err);
            throw err;
        }
    }
    checkVariables() {
        this.vars.forEach((tracker, name) => {
            // let shorten = (l: RichToken[]) => l.map((rt) => rt.token)
            // console.log(name, shorten(tracker.define), shorten(tracker.write), shorten(tracker.read))
            // console.log('')
            if (tracker.declare.length) {
                return;
            }
            if (tracker.define.length > 1) {
                tracker.define.forEach(t => t.error = lint_1.getError(''));
            }
            let writes = tracker.define.concat(tracker.write);
            let reads = tracker.read.concat(tracker.delete, tracker.exists);
            if (!writes.length) {
                reads.forEach(rT => rT.warning = lint_1.getError('unwritten', undefined, rT));
            }
            else if (!reads.length) {
                writes.forEach(rT => rT.warning = lint_1.getError('unread', undefined, rT));
            }
        });
    }
    checkFunctions() {
        this.funcs.forEach((tracker, name) => {
            if (tracker.func.length > 1) {
                tracker.func.forEach(rT => rT.error = lint_1.getError('multidef', undefined, rT));
            }
            else if (!tracker.func.length) {
                tracker.call.forEach(rT => rT.error = lint_1.getError('nodef', undefined, rT));
            }
        });
    }
    getVar(name) {
        let result = this.vars.get(name);
        if (!result) {
            result = { read: [], write: [], exists: [], delete: [], define: [], declare: [] };
            this.vars.set(name, result);
        }
        return helpers_1.surely(result);
    }
    getFunc(name) {
        let result = this.funcs.get(name);
        if (!result) {
            result = { func: [], call: [] };
            this.funcs.set(name, result);
        }
        return helpers_1.surely(result);
    }
    name() { return this.doc.uri.fsPath.replace(/.*[\\/]/, ''); }
    dir() { return this.doc.uri.fsPath.replace(/([\\/])[^\\/]*$/, '$1'); }
    updateDoc() { updateDoc(this.doc); }
    getWord(position) {
        let wordRange = this.doc.getWordRangeAtPosition(position, tokenPattern);
        let word = wordRange && this.doc.getText(wordRange);
        let tokenMatch = wordRange && this.tokens.flat.find((rToken) => {
            let intersection = wordRange.intersection(rToken.range);
            return !!(intersection && !intersection.isEmpty);
        });
        return { wordRange, word, tokenMatch };
    }
    diagnose(diagnostics) {
        let result = [];
        result.push(...this.tokens.flat
            .filter(rT => {
            return rT.error && !rT.meta.notLinting
                && !(rT.meta.disabledLints && rT.meta.disabledLints.has(rT.error.code));
        })
            .map(rT => {
            let error = helpers_1.surely(rT.error);
            let diagnostic = new vscode.Diagnostic(rT.range, error.message);
            diagnostic.code = error.code;
            diagnostic.source = 'crpl';
            return diagnostic;
        }), ...this.tokens.flat
            .filter(rT => {
            return rT.warning && !rT.meta.notLinting
                && !(rT.meta.disabledLints && rT.meta.disabledLints.has(rT.warning.code));
        })
            .map(rT => {
            let warning = helpers_1.surely(rT.warning);
            let diagnostic = new vscode.Diagnostic(rT.range, warning.message, vscode.DiagnosticSeverity.Warning);
            diagnostic.code = warning.code;
            diagnostic.source = 'crpl';
            return diagnostic;
        }));
        diagnostics.set(this.doc.uri, result);
    }
    printTokens(f = t => t.token || (t.id ? `<${t.id}>` : t)) {
        class StringTree extends Array {
        }
        function niceTree(branch) {
            if (Array.isArray(branch)) {
                return branch.map(niceTree);
            }
            else {
                return f(branch) ? f(branch) : branch;
            }
        }
        console.log({
            lines: this.tokens.lines.map(line => line.map(rToken => rToken.token).join(', ')),
            tree: this.tokens.tree.map(niceTree)
        });
    }
}
RichDoc.valuePattern = /"|-?\d+(\.\d*)?/;
RichDoc.keyofVarTracker = /^(read|write|exists|delete)$/;
RichDoc.keyofFuncTracker = /^(func|call)$/;
RichDoc.stackSpec = helpers_1.stackSpec;
function docuWikiDocToMD(crplHTML) {
    try {
        // tidy up
        let docu = crplHTML.replace(/\r/g, '');
        docu = /=====(.*\n)*/.exec(docu)[0];
        docu = docu.replace(/(===+\s*Examples|<\/textarea)(.*\n)*/, '');
        docu = docu.replace(/''|%%/g, '`');
        docu = entityDecode(docu);
        let splitKey = `~~~ ${Math.random()} ~~~`;
        let matchedResult = docPattern.exec(docu);
        if (matchedResult === null) {
            throw new Error('400');
        }
        let [, id, , args, results, notation, , description] = matchedResult;
        description = description.replace(/==+ *(.*?) *==+/g, `\n${splitKey}\n**$1**\n`)
            .replace(/\[\[.*\|(.*)\]\]/g, '$1')
            .replace(/<\/?note>/g, '\n')
            .replace(/<code>/g, '```crpl\n')
            .replace(/<\/code>/g, '\n```');
        // replace docuWiki table with MD table
        let newDescription = description.replace(/\^(.*)\n/g, '|$1\n|\n');
        while (newDescription !== description) {
            description = newDescription;
            newDescription = description.replace(/\^(.*)\n/, '|$1\n|---');
        }
        return [
            `\`${id}\`: ${notation}`,
            `[${args}] - [${results}]`,
            ...description.split(splitKey)
        ];
    }
    catch (err) {
        console.error(err);
        throw err;
    }
}
function completionFilter(completionList, rToken, pattern, prefix) {
    let word = rToken.token;
    let niceWord = word.toLowerCase().replace(pattern, '');
    let result = completionList
        .filter(s => pattern.test(s))
        .map(s => s.replace(pattern, ''))
        .filter(completion => completion.toLowerCase().startsWith(niceWord))
        .map(s => prefix + s);
    return result.map(c => ({
        label: c, range: rToken.range
    }));
}
function toggleSyntax(editor) {
    let doc = editor.document;
    let config = vscode.workspace.getConfiguration('crpl', doc.uri);
    let live = config.get('live');
    config.update('live', !live);
}
function updateDoc(doc) {
    if (vscode.languages.match(crplSelector, doc)) {
        let rDoc = documents.get(doc.uri.fsPath);
        if (!rDoc) {
            rDoc = new RichDoc(doc);
            documents.set(doc.uri.fsPath, rDoc);
        }
        else if (rDoc.doc.isClosed) {
            rDoc.doc = doc;
        }
        rDoc.tokenize();
        rDoc.diagnose(diagnostics);
    }
}
function updateDocMaybe(doc) {
    let config = vscode.workspace.getConfiguration('crpl', doc.uri);
    if (config.get('live')) {
        updateDoc(doc);
    }
}
function activate(context) {
    if (false) {
        helpers_1.scrapeSignatures();
    }
    vscode.workspace.textDocuments.forEach(updateDoc);
    context.subscriptions.push(vscode.workspace.onDidOpenTextDocument(updateDoc), vscode.workspace.onDidChangeTextDocument(change => { updateDocMaybe(change.document); }), vscode.workspace.onDidSaveTextDocument(updateDoc), vscode.workspace.onDidChangeConfiguration(() => vscode.workspace.textDocuments.forEach(updateDoc)), vscode.workspace.onDidCloseTextDocument(doc => documents.delete(doc.uri.fsPath)), vscode.languages.registerHoverProvider(crplSelector, {
        provideHover(document, position, token) {
            return __awaiter(this, void 0, void 0, function* () {
                let rDoc = documents.get(document.uri.fsPath);
                let { wordRange, word, tokenMatch } = rDoc.getWord(position);
                if (word) {
                    try {
                        if (tokenMatch && tokenMatch.meta.wiki && !tokenMatch.error && tokenMatch.id !== undefined) {
                            if (tokenMatch.id.startsWith('const_')) {
                                return new vscode.Hover(`\`${tokenMatch.id.toUpperCase()}\`: ${crplData.unitConstants[tokenMatch.id]}`, wordRange);
                            }
                            let docs = yield request(crplData.prefix + tokenMatch.id + crplData.suffix);
                            return new vscode.Hover(docuWikiDocToMD(docs), wordRange);
                        }
                        else {
                            return undefined;
                        }
                    }
                    catch (err) {
                        if (err === 400) {
                            return new vscode.Hover(`Unable to parse response from ${crplData.prefix + tokenMatch.id + crplData.suffix}.`);
                        }
                        else {
                            console.error(err);
                        }
                    }
                }
            });
        }
    }), vscode.languages.registerCompletionItemProvider(crplSelector, {
        provideCompletionItems(document, position, token) {
            return __awaiter(this, void 0, void 0, function* () {
                let rDoc = documents.get(document.uri.fsPath);
                let { wordRange, word, tokenMatch } = rDoc.getWord(position.with(undefined, position.character - 1));
                if (word && tokenMatch && wordRange.end.isEqual(position)) {
                    let pattern = /^/;
                    prefixPatterns.forEach(p => {
                        if (p.test(word)) {
                            pattern = p;
                        }
                    });
                    let prefix = pattern.exec(word)[0];
                    let longlist = completionFilter(rDoc.tokens.flat.filter(rt => tokenMatch !== rt).map(rt => rt.token), tokenMatch, pattern, prefix);
                    longlist.push(...completionFilter(crplData.completionList, tokenMatch, pattern, prefix));
                    let dupelog = [];
                    let result = [];
                    longlist.forEach((c, i) => {
                        if (dupelog.indexOf(c.label) === -1) {
                            dupelog.push(c.label);
                            result.push(c);
                        }
                    });
                    return result;
                }
            });
        }
    }), vscode.languages.registerDefinitionProvider(crplSelector, {
        provideDefinition(document, position, token) {
            return __awaiter(this, void 0, void 0, function* () {
                let rDoc = documents.get(document.uri.fsPath);
                let { tokenMatch } = rDoc.getWord(position);
                if (tokenMatch && (tokenMatch.meta.var)) {
                    let tracker = rDoc.getVar(tokenMatch.meta.var);
                    let definition = tracker.define[0] || tracker.write[0];
                    if (definition) {
                        return new vscode.Location(rDoc.doc.uri, definition.range);
                    }
                }
                else if (tokenMatch && tokenMatch.meta.func) {
                    let definition = rDoc.getFunc(tokenMatch.meta.func).func[0];
                    if (definition) {
                        return new vscode.Location(rDoc.doc.uri, definition.range);
                    }
                }
            });
        }
    }), vscode.languages.registerReferenceProvider(crplSelector, {
        provideReferences(document, position, token) {
            return __awaiter(this, void 0, void 0, function* () {
                let rDoc = documents.get(document.uri.fsPath);
                let { tokenMatch } = rDoc.getWord(position);
                if (tokenMatch && (tokenMatch.meta.var)) {
                    let tracker = rDoc.getVar(tokenMatch.meta.var);
                    return [...tracker.define, ...tracker.write, ...tracker.read, ...tracker.delete, ...tracker.exists, ...tracker.declare]
                        .map(rt => new vscode.Location(rDoc.doc.uri, rt.range));
                }
                else if (tokenMatch && tokenMatch.meta.func) {
                    let tracker = rDoc.getFunc(tokenMatch.meta.func);
                    return [...tracker.func, ...tracker.call]
                        .map(rt => new vscode.Location(rDoc.doc.uri, rt.range));
                }
            });
        }
    }), vscode.languages.registerRenameProvider(crplSelector, {
        prepareRename(document, position, token) {
            return __awaiter(this, void 0, void 0, function* () {
                let rDoc = documents.get(document.uri.fsPath);
                let { tokenMatch } = rDoc.getWord(position);
                let name = tokenMatch && (tokenMatch.meta.var || tokenMatch.meta.func);
                if (name) {
                    return {
                        range: tokenMatch.range,
                        placeholder: name
                    };
                }
            });
        },
        provideRenameEdits(document, position, newName, token) {
            return __awaiter(this, void 0, void 0, function* () {
                let rDoc = documents.get(document.uri.fsPath);
                let { tokenMatch } = rDoc.getWord(position);
                let edit = new vscode.WorkspaceEdit();
                if (tokenMatch && tokenMatch.meta.var) {
                    let name = tokenMatch.meta.var;
                    let tracker = rDoc.getVar(name);
                    [...tracker.define, ...tracker.write, ...tracker.read, ...tracker.delete, ...tracker.exists]
                        .forEach(rt => edit.replace(rDoc.doc.uri, rt.range, rt.token.replace(name, newName)));
                    return edit;
                }
                else if (tokenMatch && tokenMatch.meta.func) {
                    let name = tokenMatch.meta.func;
                    let tracker = rDoc.getFunc(name);
                    [...tracker.func, ...tracker.call]
                        .forEach(rt => edit.replace(rDoc.doc.uri, rt.range, rt.token.replace(name, newName)));
                    return edit;
                }
            });
        }
    }), vscode.commands.registerTextEditorCommand('crpl.toggleLiveSyntax', toggleSyntax));
}
exports.activate = activate;
function deactivate() {
}
exports.deactivate = deactivate;
//# sourceMappingURL=extension.js.map